#!/usr/bin/env python3

import csv
import json
import os
import random

import fiona
import numpy as np
import rasterio
from pycococreatortools import pycococreatortools
from rasterio.features import rasterize
from rasterio.windows import bounds
from shapely.geometry import box, shape

from satproc.chips import write_image
from satproc.utils import rescale_intensity, sliding_windows

output_tiles = "chips/"
output_tiles_gt = "chips_gt/"

CATEGORIES = [{"id": 1, "name": "field", "supercategory": "shape", "is_crowd": True}]

coco_output = {
    "info": {},
    "licenses": [],
    "categories": CATEGORIES,
    "images": [],
    "annotations": [],
}

COCO = "COCO"
RETINANET = "RETINANET"


def mask_from_polygons(polygons, win, mask_path, src, kwargs, image_index, mask_index):
    if polygons:
        mask = rasterize(
            polygons,
            (win.height, win.width),
            transform=rasterio.windows.transform(win, src.transform),
        )
        if mask is None:
            Exception(
                "A empty mask Was generated - Image {} Mask {}".format(
                    image_index, mask_index
                )
            )
    else:
        mask = np.zeros((win.height, win.width), dtype=np.uint8)

    # Write tile
    kwargs.update(dtype=rasterio.uint8, count=1, nodata=0)
    dst_name = "{}/{}_{}.tif".format(mask_path, image_index, mask_index)
    with rasterio.open(dst_name, "w", **kwargs) as dst:
        dst.write(mask, 1)

    return mask


def constrain_and_scale(coord, max_value, IMAGE_TILE_SIZE):
    return round((min(max(coord, 0), IMAGE_TILE_SIZE) / IMAGE_TILE_SIZE) * max_value)


def train_val_split_rows(rows, val_size=0.2):
    random.shuffle(rows)
    n_val_size = round(len(rows) * val_size)
    return rows[n_val_size:], rows[:n_val_size]


def build_dataset(
    dataset_path,
    rasters,
    chip_size=512,
    step_size=128,
    output_dir=".",
    instance=True,
    type=COCO,
    label="unknown",
    rescale_mode=None,
    rescale_range=None,
    bands=[1, 2, 3],
):

    blocks = fiona.open(dataset_path)
    block_shapes = [shape(b["geometry"]) for b in blocks]
    type = type.upper()

    k = 0
    a = 0

    if type == COCO:
        tile_path = os.path.join(output_dir, output_tiles)
        tile_gt_path = os.path.join(output_dir, output_tiles_gt)
        os.makedirs(tile_gt_path, exist_ok=True)
        os.makedirs(tile_path, exist_ok=True)

    if type == RETINANET:
        rows = []

    for raster in rasters:
        k = 0
        with rasterio.open(raster) as src:
            tile_path = os.path.join(output_dir, output_tiles, os.path.basename(raster))
            os.makedirs(tile_path, exist_ok=True)
            win_size = (chip_size, chip_size)
            win_step_size = (step_size, step_size)
            windows = list(
                sliding_windows(
                    win_size, win_step_size, src.shape[1], src.shape[0], whole=True
                )
            )

            for (win, (i, j)) in windows:
                if type == COCO:
                    # Write tif tile
                    kwargs = src.meta.copy()
                    kwargs.update(
                        {
                            "height": win.height,
                            "width": win.width,
                            "transform": rasterio.windows.transform(win, src.transform),
                        }
                    )
                    dst_name = "{}/{}.tif".format(tile_path, k)
                    with rasterio.open(dst_name, "w", **kwargs) as dst:
                        dst.write(src.read(window=win))

                    # Append tile info on COCO dataset

                    image_info = pycococreatortools.create_image_info(
                        k, os.path.basename(dst_name), (win.height, win.width)
                    )
                    coco_output["images"].append(image_info)

                    # Get intersecting shapes with current window
                    bbox_shape = box(*bounds(win, src.transform))
                    intersect_polys = [
                        bbox_shape.intersection(s)
                        for s in block_shapes
                        if bbox_shape.intersects(s)
                    ]

                    if len(intersect_polys) > 0:
                        if instance:
                            # For each polygon, create a mask
                            for i, poly in enumerate(intersect_polys):
                                mask = mask_from_polygons(
                                    [poly], win, tile_gt_path, src, kwargs, k, i
                                )

                                if mask is not None and type == COCO:
                                    # Append annotation info on COCO dataset
                                    annotation_info = (
                                        pycococreatortools.create_annotation_info(
                                            a,
                                            k,
                                            CATEGORIES[0],
                                            mask,
                                            (512, 512),
                                            tolerance=2,
                                        )
                                    )
                                    a = a + 1
                                    coco_output["annotations"].append(annotation_info)
                        else:
                            mask = mask_from_polygons(
                                intersect_polys, win, tile_gt_path, src, kwargs, k, 0
                            )

                            if mask is not None and type == COCO:
                                # Append annotation info on COCO dataset
                                annotation_info = (
                                    pycococreatortools.create_annotation_info(
                                        a,
                                        k,
                                        CATEGORIES[0],
                                        mask,
                                        (512, 512),
                                        tolerance=2,
                                    )
                                )
                                a = a + 1
                                coco_output["annotations"].append(annotation_info)
                    else:
                        # If there aren't any polygon, an empty mask is saved
                        mask = mask_from_polygons(
                            None, win, tile_gt_path, src, kwargs, k, i
                        )
                elif type == RETINANET:
                    img = src.read(window=win)
                    img = np.nan_to_num(img)
                    img = np.array([img[b - 1, :, :] for b in bands])

                    if rescale_mode:
                        img = rescale_intensity(img, rescale_mode, rescale_range)

                    dst_name = "{}/{}.jpg".format(tile_path, k)
                    write_image(img, dst_name)

                    # Generate segments
                    segments = []
                    win_bounds = rasterio.windows.bounds(win, src.transform)
                    window_box = box(*win_bounds)

                    hits = [
                        hit
                        for _, hit in blocks.items(
                            bbox=(
                                win_bounds[0],
                                win_bounds[1],
                                win_bounds[2],
                                win_bounds[3],
                            )
                        )
                    ]
                    for hit in hits:
                        hit_shape = shape(hit["geometry"])
                        bbox = box(*hit_shape.bounds)

                        inter_bbox = window_box.intersection(bbox)
                        inter_bbox_bounds = inter_bbox.bounds
                        win_transform = rasterio.windows.transform(win, src.transform)
                        minx, maxy = ~win_transform * (
                            inter_bbox_bounds[0],
                            inter_bbox_bounds[1],
                        )
                        maxx, miny = ~win_transform * (
                            inter_bbox_bounds[2],
                            inter_bbox_bounds[3],
                        )
                        segment = dict(
                            x=minx,  # - index[0],
                            y=miny,  # - index[1],
                            width=round(maxx - minx),
                            height=round(maxy - miny),
                            label=label,
                        )
                        if segment["width"] > 0 and segment["height"] > 0:
                            segments.append(segment)

                    # Generate CSVs
                    w = chip_size
                    h = chip_size
                    for s in segments:
                        row = {}
                        x1, x2 = sorted([s["x"], s["x"] + s["width"]])
                        y1, y2 = sorted([s["y"], s["y"] + s["height"]])
                        row["x1"] = constrain_and_scale(x1, w, chip_size)
                        row["x2"] = constrain_and_scale(x2, w, chip_size)
                        row["y1"] = constrain_and_scale(y1, h, chip_size)
                        row["y2"] = constrain_and_scale(y2, h, chip_size)
                        row["tile_path"] = "{}.jpg".format(
                            os.path.join(output_tiles, os.path.basename(raster), str(k))
                        )
                        row["label"] = s["label"]
                        rows.append(row)
                else:
                    pass
                k = k + 1

    if type == COCO:
        with open(os.path.join(output_dir, "annotations_coco.json"), "w") as outfile:
            json.dump(coco_output, outfile)

    if type == RETINANET:
        rows_train, rows_val = train_val_split_rows(rows)
        for name, rows in zip(["train", "val"], [rows_train, rows_val]):
            file_path = os.path.join(output_dir, "{}.csv".format(name))
            with open(file_path, "w") as csvfile:
                writer = csv.DictWriter(
                    csvfile, fieldnames=("tile_path", "x1", "y1", "x2", "y2", "label")
                )
                for row in rows:
                    writer.writerow(row)
        with open(os.path.join(output_dir, "classes.csv"), "w") as csvfile:
            writer = csv.DictWriter(csvfile, fieldnames=("label", "index"))
            writer.writerow({"label": label, "index": 0})



import logging
import os
import tempfile
from functools import partial
from glob import glob

from tqdm import tqdm

from satproc.utils import grouper, map_with_threads, run_command

_logger = logging.getLogger(__name__)


def gdal_polygonize(src, dst):
    run_command(f"gdal_polygonize.py {src} {dst}")

def apply_threshold(src, dst, *, threshold):
    """
    Output source values (probabilities) instead of simply a binary mask

    Make sure nodata=0, so that gdal_polygonize step ignores pixels under
    threshold.

    """
    # Rescale to 255
    threshold = threshold * 255

    run_command('gdal_calc.py '
                f'--calc "(A >= {threshold}) * A" '
                f'-A {src} '
                '--NoDataValue 0 '
                f'--outfile {dst}')
    

def process_image(image, *, tmpdir, threshold):
    src = image
    if threshold:
        src = os.path.join(tmpdir, os.path.basename(image))
        apply_threshold(src=image, dst=src, threshold=threshold)
    name, _ = os.path.splitext(os.path.basename(image))
    dst = os.path.join(tmpdir, f'{name}.gpkg')
    gdal_polygonize(src, dst)




def merge_vector_files(*, input_dir, output, tmpdir):
    srcs = list(glob(os.path.join(input_dir, '*.gpkg')))
    src_groups = list(enumerate(grouper(srcs, n=1000)))
    groups_dir = os.path.join(tmpdir, 'groups')
    os.makedirs(groups_dir, exist_ok=True)

    def merge_chip_vector_files(enumerated_srcs, *, output_dir):
        i, srcs = enumerated_srcs
        srcs = [f for f in srcs if f]
        output = os.path.join(groups_dir, f'{i}.gpkg')
        run_command(
            f'ogrmerge.py -overwrite_ds -single '
            f'-f GPKG -o {output} {" ".join(srcs)}',
            quiet=False)
        return output

    # First, merge groups of vector files using ogrmerge.py in parallel
    output_dir = os.path.join(tmpdir, 'temp')
    worker = partial(merge_chip_vector_files, output_dir=output_dir)
    map_with_threads(src_groups, worker)

    # Second, merge ogrmerge results using ogr2ogr into a single file
    group_paths = glob(os.path.join(groups_dir, '*.gpkg'))
    os.makedirs(os.path.dirname(output), exist_ok=True)
    for src in tqdm(group_paths):
        run_command(f'ogr2ogr -f GPKG -update -append {output} {src}',
                    quiet=False)



def retile_all(input_files, tile_size, temp_dir):
    tiles_dir = os.path.join(temp_dir, "_tiles")
    for raster in input_files:
        retile(raster, output_dir=tiles_dir, tile_size=tile_size)
    tiles_files = list(glob(os.path.join(tiles_dir, "*.tif")))
    _logger.info(
        "Num. tiles sized %dx%d at %s: %d",
        tile_size,
        tile_size,
        tiles_dir,
        len(tiles_files),
    )
    return tiles_files


def retile(raster, output_dir, tile_size):
    os.makedirs(output_dir, exist_ok=True)
    run_command(
        "gdal_retile.py "
        f"-ps {tile_size} {tile_size} "
        f"-targetDir {output_dir} {raster}"
    )


def polygonize(threshold=None, temp_dir=None, *, input_dir, output):
    images = list(glob(os.path.join(input_dir, "*.tif")))

    must_remove_temp_dir = False
    if temp_dir:
        # Make sure directory exists
        os.makedirs(temp_dir, exist_ok=True)
    else:
        # Create a temporary directory if user did not provide one
        must_remove_temp_dir = True
        tmpdir = tempfile.TemporaryDirectory()
        temp_dir = tmpdir.name

    # Process all chip images
    worker = partial(process_image, tmpdir=temp_dir, threshold=threshold)
    map_with_threads(images, worker)

    # Merge all vector files into a single one
    merge_vector_files(input_dir=temp_dir, output=output, tmpdir=temp_dir)

    if must_remove_temp_dir:
        tmpdir.cleanup()



import logging
import os

# Workaround: Load fiona at the end to avoid segfault on box (???)
import fiona
import numpy as np
import rasterio
from rasterio.crs import CRS
from rasterio.features import rasterize
from shapely.geometry import box, shape
from shapely.ops import unary_union
from skimage import exposure
from skimage.io import imsave
from tqdm import tqdm

from satproc.utils import rescale_intensity, sliding_windows, write_chips_geojson

__author__ = "Damián Silvani"
__copyright__ = "Dymaxion Labs"
__license__ = "Apache-2.0"

_logger = logging.getLogger(__name__)


def get_shape(feature):
    """Get shape geometry from feature

    Parameters
    ----------
    feature : dict
        Feature as read from Fiona

    Returns
    -------
    shapely.geometry.BaseGeometry

    """
    geom = feature["geometry"]
    try:
        return shape(geom)
    except Exception as err:
        _logger.warn("Failed to get shape from feature %s: %s", feature, err)
        return None


def mask_from_polygons(polygons, *, win, t):
    """Generate a binary mask array from a set of polygon

    Parameters
    ----------
    polygons : List[Union[Polygon, MultiPolygon]]
        list of polygon or multipolygon geometries
    win : rasterio.windows.Window
        window
    t : rasterio.transform.Affine
        affine transform

    Returns
    -------
    numpy.ndarray

    """
    transform = rasterio.windows.transform(win, t)
    if polygons is None or len(polygons) == 0:
        mask = np.zeros((win.height, win.width), dtype=np.uint8)
    else:
        mask = rasterize(
            polygons, (win.height, win.width), default_value=255, transform=transform
        )
    return mask


def multiband_chip_mask_by_classes(
    classes,
    transform,
    window,
    mask_path,
    label_property,
    polys_dict=None,
    label_path=None,
    window_shape=None,
    metadata={},
):
    multi_band_mask = []
    if polys_dict is None and label_path is not None:
        polys_dict = classify_polygons(label_path, label_property, classes)
    if window_shape is None:
        window_shape = box(*rasterio.windows.bounds(window, transform))

    for k in classes:
        multi_band_mask.append(
            mask_from_polygons(polys_dict[k], win=window, t=transform)
        )

    kwargs = metadata.copy()
    kwargs.update(
        driver="GTiff",
        dtype=rasterio.uint8,
        count=len(multi_band_mask),
        nodata=0,
        transform=rasterio.windows.transform(window, transform),
        width=window.width,
        height=window.height,
    )

    os.makedirs(os.path.dirname(mask_path), exist_ok=True)
    with rasterio.open(mask_path, "w", **kwargs) as dst:
        for i in range(len(multi_band_mask)):
            dst.write(multi_band_mask[i], i + 1)


def classify_polygons(labels, label_property, classes):
    with fiona.open(labels) as blocks:
        polys_dict = {}
        for block in blocks:
            if label_property in block["properties"]:
                c = str(block["properties"][label_property])
                try:
                    geom = shape(block["geometry"])
                except RuntimeError:
                    _logger.warning(
                        "Failed to get geometry shape for feature: %s", block
                    )
                    continue
                if c in polys_dict:
                    polys_dict[c].append(geom)
                else:
                    polys_dict[c] = [geom]
    if classes:
        for c in classes:
            if c not in polys_dict:
                polys_dict[c] = []
                _logger.warn(
                    f"No features of class '{c}' found. Will generate empty masks."
                )
    return polys_dict


def prepare_aoi_shape(aoi):
    with fiona.open(aoi) as src:
        aoi_polys = [get_shape(f) for f in src]
        aoi_polys = [shp for shp in aoi_polys if shp and shp.is_valid]
        aoi_poly = unary_union(aoi_polys)
        return aoi_poly


def prepare_label_shapes(
    labels, mask_type="class", label_property="class", classes=None
):
    if mask_type == "class":
        polys_dict = classify_polygons(labels, label_property, classes)
        return polys_dict
    else:
        raise RuntimeError(f"mask type '{mask_type}' not supported")


def extract_chips(
    rasters,
    aoi=None,
    labels=None,
    label_property="class",
    mask_type="class",
    rescale_mode=None,
    rescale_range=None,
    bands=None,
    type="tif",
    write_geojson=False,
    classes=None,
    crs=None,
    skip_existing=True,
    within=False,
    *,
    size,
    step_size,
    output_dir,
):
    if aoi:
        _logger.info("Prepare AOI shape")
        aoi_poly = prepare_aoi_shape(aoi)
    else:
        aoi_poly = None

    if labels:
        _logger.info("Prepare label shapes")
        polys_dict = prepare_label_shapes(
            labels, mask_type=mask_type, label_property=label_property, classes=classes
        )
    else:
        polys_dict = None

    for raster in tqdm(rasters):
        extract_chips_from_raster(
            raster,
            size=size,
            step_size=step_size,
            rescale_mode=rescale_mode,
            rescale_range=rescale_range,
            bands=bands,
            output_dir=output_dir,
            type=type,
            within=within,
            write_geojson=write_geojson,
            crs=crs,
            labels=labels,
            label_property=label_property,
            classes=classes,
            mask_type=mask_type,
            aoi_poly=aoi_poly,
            polys_dict=polys_dict,
            skip_existing=skip_existing,
        )


def extract_chips_from_raster(
    raster,
    rescale_mode=None,
    rescale_range=None,
    bands=None,
    type="tif",
    write_geojson=False,
    labels=None,
    label_property="class",
    mask_type="class",
    classes=None,
    crs=None,
    skip_existing=True,
    within=False,
    aoi_poly=None,
    polys_dict=None,
    *,
    size,
    step_size,
    output_dir,
):

    if skip_existing:
        _logger.info("Will skip existing files")

    basename, _ = os.path.splitext(os.path.basename(raster))

    masks_folder = os.path.join(output_dir, "masks")
    image_folder = os.path.join(output_dir, "images")

    with rasterio.open(raster) as ds:
        _logger.info("Raster size: %s", (ds.width, ds.height))

        if any(b > ds.count for b in bands):
            raise RuntimeError(
                f"Raster has {ds.count} bands, "
                f"but you asked to use {bands} band indexes"
            )

        if bands is None:
            bands = list(range(1, min(ds.count, 3) + 1))

        _logger.info("Building windows")
        win_size = (size, size)
        win_step_size = (step_size, step_size)
        windows = list(
            sliding_windows(win_size, win_step_size, ds.width, ds.height, whole=True)
        )
        _logger.info("Total windows: %d", len(windows))

        _logger.info("Building window shapes")
        window_shapes = [
            box(*rasterio.windows.bounds(w, ds.transform)) for w, _ in windows
        ]
        window_and_shapes = zip(windows, window_shapes)

        # Filter windows by AOI shape
        if aoi_poly:
            _logger.info("Filtering windows by AOI")
            _logger.info('Using "%s" function', "within" if within else "intersects")

            def filter_fn(w, aoi):
                if within:
                    return w.within(aoi)
                else:
                    return w.intersects(aoi)

            window_and_shapes = [
                (w, s) for w, s in window_and_shapes if filter_fn(s, aoi_poly)
            ]
            _logger.info("Total windows after filtering: %d", len(window_and_shapes))

        meta = ds.meta.copy()
        if crs:
            meta["crs"] = CRS.from_string(crs)
        if rescale_mode:
            # If rescaling, set nodata=0 (will rescale to uint8 1-255)
            meta["nodata"] = 0

        chips = []
        for c, ((window, (i, j)), win_shape) in tqdm(
            list(enumerate(window_and_shapes))
        ):
            _logger.debug("%s %s", window, (i, j))

            img_path = os.path.join(image_folder, f"{basename}_{i}_{j}.{type}")
            mask_path = os.path.join(masks_folder, f"{basename}_{i}_{j}.{type}")
            if (
                skip_existing
                and os.path.exists(img_path)
                and (not labels or os.path.exists(mask_path))
            ):
                continue

            img = ds.read(window=window)
            img = np.nan_to_num(img)
            img = np.array([img[b - 1, :, :] for b in bands])

            if rescale_mode:
                img = rescale_intensity(img, rescale_mode, rescale_range)

            if type == "tif":
                image_was_saved = write_tif(
                    img,
                    img_path,
                    window=window,
                    meta=meta.copy(),
                    transform=ds.transform,
                    bands=bands,
                )
            else:
                image_was_saved = write_image(img, img_path)

            if image_was_saved:
                chip = (win_shape, (c, i, j))
                chips.append(chip)

                if labels:
                    if mask_type == "class":
                        keys = classes if classes is not None else polys_dict.keys()
                        multiband_chip_mask_by_classes(
                            classes=keys,
                            transform=ds.transform,
                            window=window,
                            window_shape=win_shape,
                            polys_dict=polys_dict,
                            metadata=meta,
                            mask_path=mask_path,
                            label_property=label_property,
                        )

        if write_geojson:
            geojson_path = os.path.join(output_dir, "{}.geojson".format(basename))
            _logger.info("Write chips geojson at %s", geojson_path)
            write_chips_geojson(
                geojson_path, chips, type=type, crs=str(meta["crs"]), basename=basename
            )


def write_image(img, path, percentiles=None):
    rgb = np.dstack(img[:3, :, :]).astype(np.uint8)
    if exposure.is_low_contrast(rgb):
        return False
    os.makedirs(os.path.dirname(path), exist_ok=True)
    if not os.path.exists(path):
        imsave(path, rgb)
    return True


def write_tif(img, path, *, window, meta, transform, bands):
    if exposure.is_low_contrast(img):
        return False
    os.makedirs(os.path.dirname(path), exist_ok=True)
    meta.update(
        {
            "driver": "GTiff",
            "dtype": img.dtype,
            "height": window.height,
            "width": window.width,
            "transform": rasterio.windows.transform(window, transform),
            "count": len(bands),
        }
    )
    img = np.array([img[b - 1, :, :] for b in bands])
    with rasterio.open(path, "w", **meta) as dst:
        dst.write(img)
    return True



import json
import logging
import math
import multiprocessing as mp
import os
import subprocess
from itertools import zip_longest
from multiprocessing.pool import ThreadPool

import numpy as np
import rasterio
from rasterio.windows import Window
from shapely.geometry import mapping
from skimage import exposure
from tqdm import tqdm

__author__ = "Damián Silvani"
__copyright__ = "Dymaxion Labs"
__license__ = "Apache-2.0"

_logger = logging.getLogger(__name__)


def grouper(iterable, n, fillvalue=None):
    "Collect data into fixed-length chunks or blocks"
    # grouper('ABCDEFG', 3, 'x') --> ABC DEF Gxx"
    args = [iter(iterable)] * n
    return zip_longest(*args, fillvalue=fillvalue)


def sliding_windows(size, step_size, width, height, whole=False):
    """Slide a window of +size+ by moving it +step_size+ pixels

    Parameters
    ----------
    size : int
        window size, in pixels
    step_size : int
        step or *stride* size when sliding window, in pixels
    width : int
        image width
    height : int
        image height
    whole : bool (default: False)
        whether to generate only whole chips, or clip them at borders if needed

    Yields
    ------
    Tuple[Window, Tuple[int, int]]
        a pair of Window and a pair of position (i, j)

    """
    w, h = size
    sw, sh = step_size
    end_i = height - h if whole else height
    end_j = width - w if whole else width
    for pos_i, i in enumerate(range(0, end_i, sh)):
        for pos_j, j in enumerate(range(0, end_j, sw)):
            real_w = w if whole else min(w, abs(width - j))
            real_h = h if whole else min(h, abs(height - i))
            yield Window(j, i, real_w, real_h), (pos_i, pos_j)


def rescale_intensity(image, rescale_mode, rescale_range):
    """
    Calculate percentiles from a range cut and
    rescale intensity of image to byte range

    Parameters
    ----------
    image : numpy.ndarray
        image array
    rescale_mode : str
        rescaling mode, either 'percentiles' or 'values'
    rescale_range : Tuple[number, number]
        input range for rescaling

    Returns
    -------
    numpy.ndarray
        rescaled image
    """

    if rescale_mode == "percentiles":
        in_range = np.percentile(image, rescale_range, axis=(1, 2)).T
    elif rescale_mode == "values":
        min_value, max_value = rescale_range
        if not min_value:
            min_value = np.min(image)
        if not max_value:
            max_value = np.max(image)
        in_range = (min_value, max_value)
    elif rescale_mode == "s2_rgb_extra":

        in_range = np.percentile(image, rescale_range, axis=(1, 2)).T
        # Override first 3 ranges for (0, 0.3) (Sentinel-2 L2A TCI range)

        in_range[0] = (0, 0.3)
        in_range[1] = (0, 0.3)
        in_range[2] = (0, 0.3)

    else:
        raise RuntimeError(f"unknown rescale_mode {rescale_mode}")

    # return exposure.rescale_intensity(
    #     image, in_range=in_range, out_range=(0, 255)
    # ).astype(np.uint8)

    return np.array(
        [
            exposure.rescale_intensity(
                image[i, :, :], in_range=tuple(in_range[i]), out_range=(1, 255)
            ).astype(np.uint8)
            for i in range(image.shape[0])
        ]
    )


def calculate_raster_percentiles(
    raster, lower_cut=2, upper_cut=98, sample_size=5000, block_size=2048
):
    """Estimate raster percentiles in blocks

    This method tiles raster image in blocks, samples block
    and calculates percentiles on all samples.

    Useful for large images that don't fit memory.

    Parameters
    ----------
    raster : str
        path to raster image
    lower_cut : float
        lower cut percentile (default: 2)
    upper_cut : float
        upper cut percentile (default: 98)
    sample_size : int
        how many samples to take for each block (default: 5000)
    block_size : int
        size of square block (default: 2048)

    Returns
    -------
    Tuple[float, float]
        raster percentiles

    """
    sample_size = 5000
    size = (2048, 2048)

    with rasterio.open(raster) as ds:
        windows = list(sliding_windows(size, size, ds.width, ds.height))
        window_sample_size = math.ceil(sample_size / len(windows))
        _logger.info(
            "Windows: %d, windows sample size: %d", len(windows), window_sample_size
        )
        totals_per_bands = [[] for _ in range(ds.count)]
        for window, _ in tqdm(windows):
            img = ds.read(window=window)
            img = np.nan_to_num(img)
            window_sample = []
            for i in range(img.shape[0]):
                values = img[i].flatten()
                window_sample.append(
                    np.random.choice(values, size=window_sample_size, replace=False)
                )
            for i, win in enumerate(window_sample):
                totals_per_bands[i].append(win)

        for i, totals in enumerate(totals_per_bands):
            totals_per_bands[i] = np.array(totals).flatten()

        totals = np.array(totals_per_bands)
        _logger.info("Total shape: %s", totals.shape)

        res = tuple(
            tuple(p) for p in np.percentile(totals, (lower_cut, upper_cut), axis=1).T
        )
        _logger.info("Percentiles: %s", res)

        return res


def write_chips_geojson(output_path, chip_pairs, *, type, crs, basename):
    """Write a GeoJSON containing chips polygons as features

    Parameters
    ----------
    output_path : str
        GeoJSON output path
    chip_pairs : Tuple[Shape, Tuple[int, int, int]]
        a pair with the chip polygon geometry, and a tuple of (feature id, x, y)
    type : str
        chip file type extension (e.g. tif, jpg)
    crs : str
        CRS epsg code of chip polygon geometry
    basename : str
        basename of chip files

    Returns
    -------
    None

    """
    if not chip_pairs:
        _logger.warn("No chips to save")
        return

    _logger.info("Write chips geojson")
    os.makedirs(os.path.dirname(output_path), exist_ok=True)

    with open(output_path, "w") as f:
        d = {"type": "FeatureCollection", "features": []}
        if crs != "epsg:4326":
            code = crs.split(":")[1]
            d["crs"] = {
                "type": "name",
                "properties": {"name": f"urn:ogc:def:crs:EPSG::{code}"},
            }
        for i, (chip, (_fi, xi, yi)) in enumerate(chip_pairs):
            filename = f"{basename}_{xi}_{yi}.{type}"
            feature = {
                "type": "Feature",
                "geometry": mapping(chip),
                "properties": {"id": i, "x": xi, "y": yi, "filename": filename},
            }
            d["features"].append(feature)
        f.write(json.dumps(d))


def get_raster_band_count(path):
    """Get raster band count

    Parameters
    ----------
    path : str
        path of the raster image

    Returns
    -------
    int
        band count

    """
    with rasterio.open(path) as src:
        return src.count


def run_command(cmd, quiet=True):
    """Run a shell command

    Parameters
    ----------
    cmd : str
        command to run
    quiet : bool (default: True)
        silent output (stdout and sterr)

    Returns
    -------
    None

    """
    stderr = subprocess.DEVNULL if quiet else None
    stdout = subprocess.DEVNULL if quiet else None
    subprocess.run(cmd, shell=True, stderr=stderr, stdout=stdout)


def map_with_threads(items, worker, num_jobs=None, total=None):
    """Map a worker function to an iterable of items, using a thread pool

    Parameters
    ----------
    items : iterable
        items to map
    worker : Function
        worker function to apply to each item
    num_jobs : int
        number of threads to use
    total : int (optional)
        total number of items (for the progress bar)

    Returns
    -------
    None

    """
    if not total:
        total = len(items)
    if not num_jobs:
        num_jobs = mp.cpu_count()
    with ThreadPool(num_jobs) as pool:
        with tqdm(total=len(items)) as pbar:
            for _ in enumerate(pool.imap_unordered(worker, items)):
                pbar.update()



# -*- coding: utf-8 -*-
"""
This is a skeleton file that can serve as a starting point for a Python
console script. To run this script uncomment the following lines in the
[options.entry_points] section in setup.cfg:

    console_scripts =
         fibonacci = satproc.skeleton:run

Then run `python setup.py install` which will install the command `fibonacci`
inside your current environment.
Besides console scripts, the header (i.e. until _logger...) of this file can
also be used as template for Python modules.

Note: This skeleton file can be safely removed if not needed!
"""

import argparse
import logging
import sys

from satproc.build_dataset import build_dataset

__author__ = "Damián Silvani"
__copyright__ = "Dymaxion Labs"
__license__ = "Apache-2.0"

_logger = logging.getLogger(__name__)


def parse_args(args):
    """Parse command line parameters

    Args:
      args ([str]): command line parameters as list of strings

    Returns:
      :obj:`argparse.Namespace`: command line parameters namespace
    """
    parser = argparse.ArgumentParser(
        description="Create dataset from a raster file",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )

    parser.add_argument("raster", nargs="+", help="input raster file")
    parser.add_argument("dataset", help="input shape or geojson file")
    parser.add_argument(
        "--size", type=int, default=512, help="size of image tiles, in pixels"
    )
    parser.add_argument(
        "--step-size", type=int, default=128, help="step size (i.e. stride), in pixels"
    )
    parser.add_argument("-o", "--output-dir", help="output dir", default=".")

    parser.add_argument(
        "--instance",
        type=bool,
        help="generate a mask from each polygon if true, "
        "else a single mask for each chip",
        default=False,
    )
    parser.add_argument(
        "--type",
        type=str,
        help="dataset output type ['COCO', 'Retinanet']",
        default="COCO",
    )

    parser.add_argument(
        "--label",
        type=str,
        help="annotation label for retinanet dataset",
        default="unknown",
    )

    parser.add_argument("-b", "--bands", nargs="+", type=int, help="RGB band indexes")

    parser.add_argument(
        "--rescale",
        dest="rescale",
        default=True,
        action="store_true",
        help="rescale intensity using percentiles (lower/upper cuts)",
    )
    parser.add_argument(
        "--no-rescale",
        dest="rescale",
        action="store_false",
        help="do not rescale intensity",
    )
    parser.add_argument(
        "--rescale-mode",
        default="percentiles",
        choices=["percentiles", "values"],
        help="choose mode of intensity rescaling",
    )

    parser.add_argument(
        "--lower-cut",
        type=float,
        default=2,
        help="(for 'percentiles' mode) lower cut of percentiles "
        "for cumulative count in intensity rescaling",
    )
    parser.add_argument(
        "--upper-cut",
        type=float,
        default=98,
        help="(for 'percentiles' mode) upper cut of percentiles "
        "for cumulative count in intensity rescaling",
    )

    parser.add_argument(
        "--min",
        type=float,
        help="(for 'values' mode) minimum value in intensity rescaling",
    )
    parser.add_argument(
        "--max",
        type=float,
        help="(for 'values' mode) maximum value in intensity rescaling",
    )

    parser.add_argument(
        "-v",
        "--verbose",
        dest="loglevel",
        help="set loglevel to INFO",
        action="store_const",
        const=logging.INFO,
    )
    parser.add_argument(
        "-vv",
        "--very-verbose",
        dest="loglevel",
        help="set loglevel to DEBUG",
        action="store_const",
        const=logging.DEBUG,
    )
    return parser.parse_args(args)


def setup_logging(loglevel):
    """Setup basic logging

    Args:
      loglevel (int): minimum loglevel for emitting messages
    """
    logformat = "[%(asctime)s] %(levelname)s:%(name)s:%(message)s"
    logging.basicConfig(
        level=loglevel, stream=sys.stdout, format=logformat, datefmt="%Y-%m-%d %H:%M:%S"
    )


def main(args):
    """Main entry point allowing external calls

    Args:
      args ([str]): command line parameter list
    """
    args = parse_args(args)
    setup_logging(args.loglevel)

    bands = [1, 2, 3] if not args.bands else args.bands

    rescale_mode = args.rescale_mode if args.rescale else None
    if rescale_mode == "percentiles":
        rescale_range = (args.lower_cut, args.upper_cut)
        _logger.info("Rescale intensity with percentiles %s", rescale_range)
    elif rescale_mode == "values":
        rescale_range = (args.min, args.max)
        _logger.info("Rescale intensity with values %s", rescale_range)
    else:
        rescale_range = None
        _logger.info("No rescale intensity")

    _logger.info("Create dataset")
    build_dataset(
        args.dataset,
        args.raster,
        chip_size=args.size,
        step_size=args.step_size,
        output_dir=args.output_dir,
        instance=args.instance,
        type=args.type,
        label=args.label,
        rescale_mode=rescale_mode,
        rescale_range=rescale_range,
        bands=bands,
    )


def run():
    """Entry point for console_scripts"""
    main(sys.argv[1:])


if __name__ == "__main__":
    run()
